import { Component, createElement } from 'react';
import invariant from 'invariant';

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

const INSTANCE_MAP = new Map();
const OBSERVER_MAP = new Map();
/**
 * Monitor element, and trigger callback when element becomes visible
 * @param element {HTMLElement}
 * @param callback {Function} Called with inView
 * @param options {Object} InterSection observer options
 * @param options.threshold {Number} Number between 0 and 1, indicating how much of the element should be visible before triggering
 * @param options.root {HTMLElement} It should have a unique id or data-intersection-id in order for the Observer to reused.
 * @param options.rootMargin {String} The CSS margin to apply to the root element.
 * @param rootId {String} Unique identifier for the root element, to enable reusing the IntersectionObserver
 */

function observe(element, callback, options = {
  threshold: 0
}, rootId) {
  const {
    root,
    rootMargin
  } = options;
  const threshold = options.threshold || 0;
  if (!element || !callback) return;
  let observerId = rootMargin ? `${threshold.toString()}_${rootMargin}` : `${threshold.toString()}`;

  if (root) {
    observerId = rootId ? `${rootId}_${observerId}` : null;
  }

  let observerInstance = observerId ? OBSERVER_MAP.get(observerId) : null;

  if (!observerInstance) {
    observerInstance = new IntersectionObserver(onChange, options);
    if (observerId) OBSERVER_MAP.set(observerId, observerInstance);
  }

  const instance = {
    callback,
    visible: false,
    options,
    observerId,
    observer: !observerId ? observerInstance : undefined
  };
  INSTANCE_MAP.set(element, instance);
  observerInstance.observe(element);
  return instance;
}
/**
 * Stop observing an element. If an element is removed from the DOM or otherwise destroyed,
 * make sure to call this method.
 * @param element {HTMLElement}
 */

function unobserve(element) {
  if (!element) return;
  const instance = INSTANCE_MAP.get(element);

  if (instance) {
    const {
      observerId,
      observer
    } = instance;
    const observerInstance = observerId ? OBSERVER_MAP.get(observerId) : observer;

    if (observerInstance) {
      observerInstance.unobserve(element);
    } // Check if we are still observing any elements with the same threshold.


    let itemsLeft = false;

    if (observerId) {
      INSTANCE_MAP.forEach((item, key) => {
        if (item && item.observerId === observerId && key !== element) {
          itemsLeft = true;
        }
      });
    }

    if (observerInstance && !itemsLeft) {
      // No more elements to observe for threshold, disconnect observer
      observerInstance.disconnect();
      OBSERVER_MAP.delete(observerId);
    } // Remove reference to element


    INSTANCE_MAP.delete(element);
  }
}

function onChange(changes) {
  changes.forEach(intersection => {
    const {
      isIntersecting,
      intersectionRatio,
      target
    } = intersection;
    const instance = INSTANCE_MAP.get(target); // Firefox can report a negative intersectionRatio when scrolling.

    if (instance && intersectionRatio >= 0) {
      const options = instance.options;
      let inView = false;

      if (Array.isArray(options.threshold)) {
        // If threshold is an array, check if any of them intersects. This just triggers the onChange event multiple times.
        inView = options.threshold.some(threshold => {
          return instance.visible ? intersectionRatio > threshold : intersectionRatio >= threshold;
        });
      } else if (options.threshold !== undefined) {
        // Trigger on 0 ratio only when not visible. This is fallback for browsers without isIntersecting support
        inView = instance.visible ? intersectionRatio > options.threshold : intersectionRatio >= options.threshold;
      }

      if (isIntersecting !== undefined) {
        // If isIntersecting is defined, ensure that the element is actually intersecting.
        // Otherwise it reports a threshold of 0
        inView = inView && isIntersecting;
      }

      instance.visible = inView;
      instance.callback(inView);
    }
  });
}

/**
 * Monitors scroll, and triggers the children function with updated props
 *
 <Observer>
 {({inView, ref}) => (
   <h1 ref={ref}>{`${inView}`}</h1>
 )}
 </Observer>
 */
class Observer extends Component {
  constructor(...args) {
    super(...args);

    _defineProperty(this, "state", {
      inView: false
    });

    _defineProperty(this, "node", null);

    _defineProperty(this, "handleNode", node => {
      if (this.node) unobserve(this.node);
      this.node = node;
      this.observeNode();
    });

    _defineProperty(this, "handleChange", inView => {
      this.setState({
        inView
      });

      if (this.props.onChange) {
        this.props.onChange(inView);
      }
    });
  }

  componentDidMount() {
    if (process.env.NODE_ENV !== 'production') {
      if (this.props.hasOwnProperty('render')) {
        console.warn(`react-intersection-observer: "render" is deprecated, and should be replaced with "children"`, this.node);
      }

      invariant(this.node, `react-intersection-observer: No DOM node found. Make sure you forward "ref" to the root DOM element you want to observe.`);
    }
  }

  componentDidUpdate(prevProps, prevState) {
    // If a IntersectionObserver option changed, reinit the observer
    if (prevProps.rootMargin !== this.props.rootMargin || prevProps.root !== this.props.root || prevProps.threshold !== this.props.threshold) {
      unobserve(this.node);
      this.observeNode();
    }

    if (prevState.inView !== this.state.inView) {
      if (this.state.inView && this.props.triggerOnce) {
        unobserve(this.node);
        this.node = null;
      }
    }
  }

  componentWillUnmount() {
    if (this.node) {
      unobserve(this.node);
      this.node = null;
    }
  }

  observeNode() {
    if (!this.node) return;
    const {
      threshold,
      root,
      rootMargin,
      rootId
    } = this.props;
    observe(this.node, this.handleChange, {
      threshold,
      root,
      rootMargin
    }, rootId);
  }

  render() {
    const {
      children,
      render,
      tag,
      triggerOnce,
      threshold,
      root,
      rootId,
      rootMargin,
      ...props
    } = this.props;
    const {
      inView
    } = this.state;
    const renderMethod = children || render;

    if (typeof renderMethod === 'function') {
      return renderMethod({
        inView,
        ref: this.handleNode
      });
    }

    return createElement(tag || 'div', {
      ref: this.handleNode,
      ...props
    }, children);
  }

}

_defineProperty(Observer, "defaultProps", {
  threshold: 0,
  triggerOnce: false
});

export default Observer;
